@@ -1,7 +1,10 @@
 /* application object */
+/* vim: set sw=2 et: */
 
 /*
  * Copyright (C) 2001 Havoc Pennington
+ * Copyright (C) 2003 Red Hat, Inc.
+ * Copyright (C) 2005-2007 Vincent Untz
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -19,10 +22,33 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
+#include <glib/gi18n-lib.h>
 #include "application.h"
 #include "private.h"
 
-#define FALLBACK_NAME _("untitled application")
+/**
+ * SECTION:application
+ * @short_description: an object representing a group of windows of the same
+ * application.
+ * @see_also: wnck_window_get_application()
+ * @stability: Unstable
+ *
+ * The #WnckApplication is a group of #WnckWindow that are all in the same
+ * application. It can be used to represent windows by applications, group
+ * windows by applications or to manipulate all windows of a particular
+ * application.
+ *
+ * A #WnckApplication is identified by the group leader of the #WnckWindow
+ * belonging to it, and new #WnckWindow are added to a #WnckApplication if and
+ * only if they have the group leader of the #WnckApplication.
+ *
+ * The #WnckApplication objects are always owned by libwnck and must not be
+ * referenced or unreferenced.
+ */
+
+#define FALLBACK_NAME _("Untitled application")
 
 static GHashTable *app_hash = NULL;
 
@@ -42,6 +68,8 @@ struct _WnckApplicationPrivate
   WnckIconCache *icon_cache;
 
   WnckWindow *icon_window;
+
+  char *startup_id;
   
   guint name_from_leader : 1; /* name is from group leader */
   guint icon_from_leader : 1;
@@ -49,6 +77,9 @@ struct _WnckApplicationPrivate
   guint need_emit_icon_changed : 1;
 };
 
+G_DEFINE_TYPE (WnckApplication, wnck_application, G_TYPE_OBJECT);
+#define WNCK_APPLICATION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), WNCK_TYPE_APPLICATION, WnckApplicationPrivate))
+
 enum {
   NAME_CHANGED,
   ICON_CHANGED,
@@ -66,56 +97,53 @@ static void wnck_application_class_init  (WnckApplicationClass *klass);
 static void wnck_application_finalize    (GObject        *object);
 
 
-static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
-GType
-wnck_application_get_type (void)
-{
-  static GType object_type = 0;
-
-  g_type_init ();
-  
-  if (!object_type)
-    {
-      static const GTypeInfo object_info =
-      {
-        sizeof (WnckApplicationClass),
-        (GBaseInitFunc) NULL,
-        (GBaseFinalizeFunc) NULL,
-        (GClassInitFunc) wnck_application_class_init,
-        NULL,           /* class_finalize */
-        NULL,           /* class_data */
-        sizeof (WnckApplication),
-        0,              /* n_preallocs */
-        (GInstanceInitFunc) wnck_application_init,
-      };
-      
-      object_type = g_type_register_static (G_TYPE_OBJECT,
-                                            "WnckApplication",
-                                            &object_info, 0);
-    }
-  
-  return object_type;
-}
-
 static void
 wnck_application_init (WnckApplication *application)
 {  
-  application->priv = g_new0 (WnckApplicationPrivate, 1);
+  application->priv = WNCK_APPLICATION_GET_PRIVATE (application);
+
+  application->priv->xwindow = None;
+  application->priv->screen = NULL;
+  application->priv->windows = NULL;
+  application->priv->pid = 0;
+  application->priv->name = NULL;
+
+  application->priv->name_window = NULL;
+
+  application->priv->icon = NULL;
+  application->priv->mini_icon = NULL;
 
   application->priv->icon_cache = _wnck_icon_cache_new ();
+  _wnck_icon_cache_set_want_fallback (application->priv->icon_cache,
+                                      FALSE);
+
+  application->priv->icon_window = NULL;
+
+  application->priv->startup_id = NULL;
+
+  application->priv->name_from_leader = FALSE;
+  application->priv->icon_from_leader = FALSE;
+
+  application->priv->need_emit_icon_changed = FALSE;
 }
 
 static void
 wnck_application_class_init (WnckApplicationClass *klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
-  
-  parent_class = g_type_class_peek_parent (klass);
+
+  g_type_class_add_private (klass, sizeof (WnckApplicationPrivate));
 
   object_class->finalize = wnck_application_finalize;
   
+  /**
+   * WnckApplication::name-changed:
+   * @app: the #WnckApplication which emitted the signal.
+   *
+   * Emitted when the name of @app changes.
+   */
   signals[NAME_CHANGED] =
     g_signal_new ("name_changed",
                   G_OBJECT_CLASS_TYPE (object_class),
@@ -125,6 +153,12 @@ wnck_application_class_init (WnckApplicationClass *klass)
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
 
+  /**
+   * WnckApplication::icon-changed:
+   * @app: the #WnckApplication which emitted the signal.
+   *
+   * Emitted when the icon of @app changes.
+   */
   signals[ICON_CHANGED] =
     g_signal_new ("icon_changed",
                   G_OBJECT_CLASS_TYPE (object_class),
@@ -142,21 +176,42 @@ wnck_application_finalize (GObject *object)
 
   application = WNCK_APPLICATION (object);
 
+  application->priv->xwindow = None;
+
+  g_list_free (application->priv->windows);
+  application->priv->windows = NULL;
+  
+  g_free (application->priv->name);
+  application->priv->name = NULL;
+
   if (application->priv->icon)
     g_object_unref (G_OBJECT (application->priv->icon));
+  application->priv->icon = NULL;
 
   if (application->priv->mini_icon)
     g_object_unref (G_OBJECT (application->priv->mini_icon));
+  application->priv->mini_icon = NULL;
   
   _wnck_icon_cache_free (application->priv->icon_cache);
+  application->priv->icon_cache = NULL;
   
-  g_free (application->priv->name);
-  
-  g_free (application->priv);
-  
-  G_OBJECT_CLASS (parent_class)->finalize (object);
+  g_free (application->priv->startup_id);
+  application->priv->startup_id = NULL;
+
+  G_OBJECT_CLASS (wnck_application_parent_class)->finalize (object);
 }
 
+/**
+ * wnck_application_get:
+ * @xwindow: the X window ID of a group leader.
+ *
+ * Gets the #WnckApplication corresponding to the group leader with @xwindow
+ * as X window ID.
+ *
+ * Return value: the #WnckApplication corresponding to @xwindow, or %NULL if
+ * there no such #WnckApplication could be found. The returned #WnckApplication
+ * is owned by libwnck and must not be referenced or unreferenced.
+ */
 WnckApplication*
 wnck_application_get (gulong xwindow)
 {
@@ -168,11 +223,11 @@ wnck_application_get (gulong xwindow)
 
 /**
  * wnck_application_get_xid:
- * @app: a #WnckApplication
+ * @app: a #WnckApplication.
  * 
- * Gets the X id of the group leader window for the app.
+ * Gets the X window ID of the group leader window for @app.
  * 
- * Return value: X id for the app
+ * Return value: the X window ID of the group leader window for @app.
  **/
 gulong
 wnck_application_get_xid (WnckApplication *app)
@@ -184,12 +239,13 @@ wnck_application_get_xid (WnckApplication *app)
 
 /**
  * wnck_application_get_windows:
- * @app: a #WnckApplication
+ * @app: a #WnckApplication.
  * 
- * Gets a list of all windows belonging to @app. The list
- * is returned by reference and should not be freed.
+ * Gets the list of #WnckWindow belonging to @app.
  * 
- * Return value: list of #WnckWindow in this app
+ * Return value: the list of #WnckWindow belonging to @app, or %NULL if the
+ * application contains no window. The list should not be modified nor freed,
+ * as it is owned by @app.
  **/
 GList*
 wnck_application_get_windows (WnckApplication *app)
@@ -199,6 +255,14 @@ wnck_application_get_windows (WnckApplication *app)
   return app->priv->windows;
 }
 
+/**
+ * wnck_application_get_n_windows:
+ * @app: a #WnckApplication.
+ * 
+ * Gets the number of #WnckWindow belonging to @app.
+ * 
+ * Return value: the number of #WnckWindow belonging to @app.
+ **/
 int
 wnck_application_get_n_windows (WnckApplication *app)
 {
@@ -209,16 +273,16 @@ wnck_application_get_n_windows (WnckApplication *app)
 
 /**
  * wnck_application_get_name:
- * @app: a #WnckApplication
+ * @app: a #WnckApplication.
  * 
- * Gets the name of an application, employing various
- * suboptimal heuristics to try to figure it out.
- * Probably GTK should have a function to allow apps to
- * set _NET_WM_NAME on the group leader as the app name,
- * and the WM spec should say that's where the app name
- * goes.
+ * Gets the name of @app. Since there is no way to properly find this name,
+ * various suboptimal heuristics are used to find it. GTK+ should probably have
+ * a function to allow applications to set the _NET_WM_NAME property on the
+ * group leader as the application name, and the <ulink
+ * url="http://standards.freedesktop.org/wm-spec/wm-spec-latest.html">EWMH</ulink>
+ * should say that this is where the application name goes.
  * 
- * Return value: name of the application
+ * Return value: the name of @app, or a fallback name if no name is available.
  **/
 const char*
 wnck_application_get_name (WnckApplication *app)
@@ -235,10 +299,12 @@ wnck_application_get_name (WnckApplication *app)
  * wnck_application_get_icon_name:
  * @app: a #WnckApplication
  * 
- * Gets the icon name of an application, employing various
- * suboptimal heuristics to try to figure it out.
+ * Gets the icon name of @app (to be used when @app is minimized). Since
+ * there is no way to properly find this name, various suboptimal heuristics
+ * are used to find it.
  * 
- * Return value: name of the application when minimized
+ * Return value: the icon name of @app, or a fallback icon name if no icon name
+ * is available.
  **/
 const char*
 wnck_application_get_icon_name (WnckApplication *app)
@@ -257,12 +323,11 @@ wnck_application_get_icon_name (WnckApplication *app)
 
 /**
  * wnck_application_get_pid:
- * @app: a #WnckApplication
+ * @app: a #WnckApplication.
  * 
- * Gets the process ID of an application, or 0 if none
- * is available.
+ * Gets the process ID of @app.
  * 
- * Return value: process ID or 0
+ * Return value: the process ID of @app, or 0 if none is available.
  **/
 int
 wnck_application_get_pid (WnckApplication *app)
@@ -291,11 +356,6 @@ get_icons (WnckApplication *app)
     {
       app->priv->need_emit_icon_changed = TRUE;
       app->priv->icon_from_leader = TRUE;
-      
-      if (icon)
-        g_object_ref (G_OBJECT (icon));
-      if (mini_icon)
-        g_object_ref (G_OBJECT (mini_icon));
 
       if (app->priv->icon)
         g_object_unref (G_OBJECT (app->priv->icon));
@@ -317,34 +377,102 @@ get_icons (WnckApplication *app)
             !(app->priv->icon || app->priv->mini_icon));
 }
 
+/* Prefer to get group icon from a window of type "normal" */
+static WnckWindow*
+find_icon_window (WnckApplication *app)
+{
+  GList *tmp;
+
+  tmp = app->priv->windows;
+  while (tmp != NULL)
+    {
+      WnckWindow *w = tmp->data;
+
+      if (wnck_window_get_window_type (w) == WNCK_WINDOW_NORMAL)
+        return w;
+      
+      tmp = tmp->next;
+    }
+
+  if (app->priv->windows)
+    return app->priv->windows->data;
+  else
+    return NULL;
+}
+
+/**
+ * wnck_application_get_icon:
+ * @app: a #WnckApplication.
+ * 
+ * Gets the icon to be used for @app. If no icon is set for @app, a
+ * suboptimal heuristic is used to find an appropriate icon. If no icon was
+ * found, a fallback icon is used.
+ * 
+ * Return value: the icon for @app. The caller should reference the returned
+ * <classname>GdkPixbuf</classname> if it needs to keep the icon around.
+ **/
 GdkPixbuf*
 wnck_application_get_icon (WnckApplication *app)
 {
+  g_return_val_if_fail (WNCK_IS_APPLICATION (app), NULL);
+
   get_icons (app);
   if (app->priv->need_emit_icon_changed)
     emit_icon_changed (app);
 
-  return app->priv->icon;
+  if (app->priv->icon)
+    return app->priv->icon;
+  else
+    {
+      WnckWindow *w = find_icon_window (app);
+      if (w)
+        return wnck_window_get_icon (w);
+      else
+        return NULL;
+    }
 }
 
+/**
+ * wnck_application_get_mini_icon:
+ * @app: a #WnckApplication.
+ * 
+ * Gets the mini-icon to be used for @app. If no mini-icon is set for @app,
+ * a suboptimal heuristic is used to find an appropriate icon. If no mini-icon
+ * was found, a fallback mini-icon is used.
+ * 
+ * Return value: the mini-icon for @app. The caller should reference the
+ * returned <classname>GdkPixbuf</classname> if it needs to keep the mini-icon
+ * around.
+ **/
 GdkPixbuf*
 wnck_application_get_mini_icon (WnckApplication *app)
 {
+  g_return_val_if_fail (WNCK_IS_APPLICATION (app), NULL);
+
   get_icons (app);
   if (app->priv->need_emit_icon_changed)
     emit_icon_changed (app);
 
-  return app->priv->mini_icon;
+  if (app->priv->mini_icon)
+    return app->priv->mini_icon;
+  else
+    {
+      WnckWindow *w = find_icon_window (app);
+      if (w)
+        return wnck_window_get_mini_icon (w);
+      else
+        return NULL;
+    }
 }
 
 /**
  * wnck_application_get_icon_is_fallback:
- * @application: a #WnckApplication
+ * @app: a #WnckApplication
  *
- * Checks if we are using a default fallback icon because
- * none was set on the application.
+ * Gets whether a default fallback icon is used for @app (because none
+ * was set on @app).
  * 
- * Return value: %TRUE if icon is a fallback
+ * Return value: %TRUE if the icon for @app is a fallback, %FALSE otherwise.
  **/
 gboolean
 wnck_application_get_icon_is_fallback (WnckApplication *app)
@@ -354,6 +482,25 @@ wnck_application_get_icon_is_fallback (WnckApplication *app)
   return _wnck_icon_cache_get_is_fallback (app->priv->icon_cache);
 }
 
+/**
+ * wnck_application_get_startup_id:
+ * @app: a #WnckApplication.
+ *
+ * Gets the startup sequence ID used for startup notification of @app.
+ *
+ * Return value: the startup sequence ID used for startup notification of @app,
+ * or %NULL if none is available.
+ *
+ * Since: 2.2
+ */
+const char*
+wnck_application_get_startup_id (WnckApplication *app)
+{
+  g_return_val_if_fail (WNCK_IS_APPLICATION (app), NULL);
+
+  return app->priv->startup_id;
+}
+
 /* xwindow is a group leader */
 WnckApplication*
 _wnck_application_create (Window      xwindow,
@@ -380,6 +527,9 @@ _wnck_application_create (Window      xwindow,
     application->priv->name_from_leader = TRUE;
   
   application->priv->pid = _wnck_get_pid (application->priv->xwindow);
+
+  application->priv->startup_id = _wnck_get_utf8_property (application->priv->xwindow,
+                                                           _wnck_atom_get ("_NET_STARTUP_ID"));
   
   g_hash_table_insert (app_hash, &application->priv->xwindow, application);
   
@@ -403,8 +553,6 @@ _wnck_application_destroy (WnckApplication *application)
 
   g_return_if_fail (wnck_application_get (application->priv->xwindow) == NULL);
 
-  application->priv->xwindow = None;
-  
   /* remove hash's ref on the application */
   g_object_unref (G_OBJECT (application));
 }
@@ -437,6 +585,11 @@ _wnck_application_add_window (WnckApplication *app,
   /* emits signals, so do it last */
   reset_name (app);
   update_name (app);
+
+  /* see if we're using icon from a window */
+  if (app->priv->icon == NULL ||
+      app->priv->mini_icon == NULL)
+    emit_icon_changed (app);
 }
 
 void
@@ -456,6 +609,11 @@ _wnck_application_remove_window (WnckApplication *app,
   /* emits signals, so do it last */
   reset_name (app);
   update_name (app);
+
+  /* see if we're using icon from a window */
+  if (app->priv->icon == NULL ||
+      app->priv->mini_icon == NULL)
+    emit_icon_changed (app);
 }
 
 void
@@ -492,6 +650,11 @@ _wnck_application_process_property_notify (WnckApplication *app,
                                          xevent->xproperty.atom);
       emit_icon_changed (app);
     }
+  else if (xevent->xproperty.atom ==
+           _wnck_atom_get ("_NET_STARTUP_ID"))
+    {
+      /* FIXME update startup id */
+    }
 }
 
 static void
@@ -525,6 +688,8 @@ reset_name  (WnckApplication *app)
 static void
 update_name (WnckApplication *app)
 {
+  g_assert (app->priv->name_from_leader || app->priv->name == NULL);
+
   if (app->priv->name == NULL)
     {
       /* if only one window, get name from there. If more than one and
